/*
 * Electric(tm) VLSI Design System
 *
 * File: routmimic.c
 * Arc mimic code for the wire routing aid
 * Written by: Steven M. Rubin, Electric Editor Incorporated
 *
 * Copyright (c) 1998 Electric Editor Incorporated.
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 *
 * Electric Editor Incorporated
 * 23470 Sunset Drive, Suite 108
 * Los Gatos, California 95033
 * support@electriceditor.com
 */

#include "config.h"
#if ROUTAID

#include "global.h"
#include "rout.h"

void ro_mimicdelete(ARCPROTO *typ, NODEINST **nodes, PORTPROTO **ports);
void ro_mimiccreate(ARCINST *arc);

/*
 * Entry point for mimic router.  Called each "slice".
 */
void ro_mimicstitch(void)
{
	REGISTER ARCPROTO *ap;
	REGISTER ARCINST *ai;
	NODEINST *endnode[2];
	PORTPROTO *endport[2];
	REGISTER INTBIG options, i;

	/* if an arc was deleted, and that is being mimiced, do it */
	options = ro_getoptions();
	if (ro_deletedarc != NOARCPROTO && (options&MIMICUNROUTES) != 0)
	{
		for(i=0; i<2; i++)
		{
			endnode[i] = ro_deletednodes[i];
			endport[i] = ro_deletedports[i];
		}
		ap = ro_deletedarc;
		ro_deletedarc = NOARCPROTO;
		ro_mimicdelete(ap, endnode, endport);
	}

	/* if an arc was just created, mimic that */
	if (ro_createdarc != NOARCINST)
	{
		ai = ro_createdarc;
		ro_createdarc = NOARCINST;
		ro_mimiccreate(ai);
	}
}

/*
 * Routine to mimic the unrouting of an arc that ran from nodes[0]/ports[0] to
 * nodes[1]/ports[1] with type "typ".
 */
void ro_mimicdelete(ARCPROTO *typ, NODEINST **nodes, PORTPROTO **ports)
{
	REGISTER NODEPROTO *facet;
	REGISTER ARCINST *ai, *nextai;
	REGISTER INTBIG match, dist, thisdist, deleted, angle, thisangle;
	INTBIG x0, y0, x1, y1;

	/* determine length of deleted arc */
	portposition(nodes[0], ports[0], &x0, &y0);
	portposition(nodes[1], ports[1], &x1, &y1);
	dist = computedistance(x0, y0, x1, y1);
	if (dist != 0)
		angle = figureangle(x0, y0, x1, y1);

	/* look for a similar situation to delete */
	deleted = 0;
	facet = nodes[0]->parent;
	for(ai = facet->firstarcinst; ai != NOARCINST; ai = nextai)
	{
		nextai = ai->nextarcinst;

		/* arc must be of the same type */
		if (ai->proto != typ) continue;

		/* arc must connect to the same type of node/port */
		match = 0;
		if (ai->end[0].nodeinst->proto == nodes[0]->proto &&
			ai->end[1].nodeinst->proto == nodes[1]->proto &&
			ai->end[0].portarcinst->proto == ports[0] &&
			ai->end[1].portarcinst->proto == ports[1]) match = 1;
		if (ai->end[0].nodeinst->proto == nodes[1]->proto &&
			ai->end[1].nodeinst->proto == nodes[0]->proto &&
			ai->end[0].portarcinst->proto == ports[1] &&
			ai->end[1].portarcinst->proto == ports[0]) match = -1;
		if (match == 0) continue;

		/* must be the same length and angle */
		portposition(ai->end[0].nodeinst, ai->end[0].portarcinst->proto, &x0, &y0);
		portposition(ai->end[1].nodeinst, ai->end[1].portarcinst->proto, &x1, &y1);
		thisdist = computedistance(x0, y0, x1, y1);
		if (dist != thisdist) continue;
		if (dist != 0)
		{
			thisangle = figureangle(x0, y0, x1, y1);
			if ((angle%1800) != (thisangle%1800)) continue;
		}

		/* the same! delete it */
		startobjectchange((INTBIG)ai, VARCINST);
		killarcinst(ai);
		deleted++;
	}
	if (deleted != 0)
		ttyputmsg("MIMIC ROUTING: deleted %ld %s", deleted, makeplural("wire", deleted));
}

/*
 * Routine to mimic the creation of arc "arc".
 */
void ro_mimiccreate(ARCINST *ai)
{
	REGISTER NODEPROTO *facet, *proto0, *proto1;
	REGISTER NODEINST *ni, *oni, *node0, *node1;
	REGISTER PORTPROTO *port0, *port1, *pp, *opp;
	REGISTER PORTARCINST *pi, *opi;
	REGISTER ARCINST *newai;
	REGISTER INTBIG dist, angle, options, count, i, wantx1, wanty1, end, end0offx, end0offy;
	INTBIG x0, y0, x1, y1, x0c, y0c;
	static POLYGON *poly = NOPOLYGON;

	/* get polygon */
	if (poly == NOPOLYGON) poly = allocstaticpolygon(5, ro_aid->cluster);

	facet = ai->parent;
	options = ro_getoptions();
	count = 0;

	/* search from both ends */
	for(end=0; end<2; end++)
	{
		node0 = ai->end[end].nodeinst;
		node1 = ai->end[1-end].nodeinst;
		proto0 = node0->proto;
		proto1 = node1->proto;
		port0 = ai->end[end].portarcinst->proto;
		port1 = ai->end[1-end].portarcinst->proto;
		x0 = ai->end[end].xpos;     y0 = ai->end[end].ypos;
		x1 = ai->end[1-end].xpos;   y1 = ai->end[1-end].ypos;
		dist = computedistance(x0, y0, x1, y1);
		if (dist != 0)
			angle = figureangle(x0, y0, x1, y1);
		portposition(node0, port0, &x0c, &y0c);
		end0offx = x0 - x0c;   end0offy = y0 - y0c;

		/* now search every node in the facet */
		for(ni = facet->firstnodeinst; ni != NONODEINST; ni = ni->nextnodeinst)
		{
			if (ni->proto != proto0) continue;
			if (ni->highx - ni->lowx != node0->highx - node0->lowx) continue;
			for(pp = ni->proto->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
			{
				if ((options&MIMICIGNOREPORTS) == 0)
				{
					/* considering precise ports: check it */
					if (pp != port0) continue;
				} else
				{
					/* nonspecific ports: make sure the arc can connect */
					for(i=0; pp->connects[i] != NOARCPROTO; i++)
						if (pp->connects[i] == ai->proto) break;
					if (pp->connects[i] == NOARCPROTO) continue;
				}
				portposition(ni, pp, &x0, &y0);
				x0 += end0offx;   y0 += end0offy;
				if (dist == 0)
				{
					wantx1 = x0;   wanty1 = y0;
				} else
				{
					wantx1 = x0 + mult(dist, cosine((INTSML)angle));
					wanty1 = y0 + mult(dist, sine((INTSML)angle));
				}

				/* now look for another node that matches the situation */
				for(oni = facet->firstnodeinst; oni != NONODEINST; oni = oni->nextnodeinst)
				{
					if (oni->proto != proto1) continue;
					if (oni->highx - oni->lowx != node1->highx - node1->lowx) continue;

					/* ensure that intra-node wirings stay that way */
					if (node0 == node1)
					{
						if (ni != oni) continue;
					} else
					{
						if (ni == oni) continue;
					}
					for(opp = oni->proto->firstportproto; opp != NOPORTPROTO; opp = opp->nextportproto)
					{
						if ((options&MIMICIGNOREPORTS) == 0)
						{
							/* considering precise ports: check it */
							if (opp != port1) continue;
						} else
						{
							/* nonspecific ports: make sure the arc can connect */
							for(i=0; opp->connects[i] != NOARCPROTO; i++)
								if (opp->connects[i] == ai->proto) break;
							if (opp->connects[i] == NOARCPROTO) continue;
						}

						/* don't replicate what is already done */
						if (ni == node0 && pp == port0 && oni == node1 && opp == port1) continue;

						/* see if they are the same distance apart */
						shapeportpoly(oni, opp, poly, 0);
						if (isinside(wantx1, wanty1, poly) == 0) continue;

						/* see if they are already wired */
						for(pi = ni->firstportarcinst; pi != NOPORTARCINST; pi = pi->nextportarcinst)
						{
							if (pi->proto != pp) continue;
							for (opi = oni->firstportarcinst; opi != NOPORTARCINST; opi = opi->nextportarcinst)
							{
								if (opi->proto != opp) continue;
								if (pi->conarcinst == opi->conarcinst) break;
							}
							if (opi != NOPORTARCINST) break;
						}
						if (pi != NOPORTARCINST) continue;

						newai = newarcinst(ai->proto, ai->width, ai->userbits, ni,
							pp, x0, y0, oni, opp, wantx1, wanty1, facet);
						if (newai == NOARCINST)
						{
							ttyputerr("Problem creating arc");
							return;
						}
						endobjectchange((INTBIG)newai, VARCINST);
						count++;
					}
				}
			}
		}
	}
	if (count != 0)
		ttyputmsg("MIMIC ROUTING: added %ld %s", count, makeplural("wire", count));
}

#endif  /* ROUTAID - at top */
